'#include once "fbmld.bi"
#include once "lwo/lwconstants.bi"
#include once "lwo/load_lwo.bi"
#include once "image.bi"
#include once "mathfuncf.bi"
#include once "hash.bi"

Const as single Shadow_Infinity = 1000.0

namespace nmodel

	type surface_struct
		surfName			As string
		Start_ID       as uinteger
		End_ID         as uinteger
		Shade_Mode     as uinteger
		Double_Sided   as integer
		Is_Tex			as byte
		Tex1				as integer
		Tex2		     	as integer
		Tex3		     	as integer
		Tex4		     	as integer
		Texb			 	as integer
		Is_Multi		 	as byte
		Is_Bump			as byte
		Rgbalpha       as nimage.Rgbalpha
		Ambient        as nimage.Rgbalpha
		Diffuse        as nimage.Rgbalpha
		Specular       as nimage.Rgbalpha
		Emission       as nimage.Rgbalpha
		Shininess      as single
		dList          as gluint
		shader			as glHandleARB
		Group			 	as integer
		
		has_shader		 as integer
		dim shader_callback as sub( byref surface as surface_struct, texture() as nimage.texture_struct )
		
	end type

	type triangle_struct
		point_id(1 to 3) 	as uinteger
		t1coord(1 to 3)	as nmathf.vec2f
		t2coord(1 to 3)	as nmathf.vec2f
		t3coord(1 to 3)	as nmathf.vec2f
		t4coord(1 to 3)	as nmathf.vec2f
		bcoord(1 to 3)	 	as nmathf.vec2f
		vnormals(1 to 3) 	as nmathf.vec3f
		con(1 to 3)      	as integer
		Visible          	as integer
		invisible        	as integer
		normal           	as nmathf.vec3f
		Plane            	as nmathf.vec4f
		Group            	as integer
		string_id        	as string
		flagged				as integer
	end type

	type bone_struct
		m					as nmathf.matrix
		weight(any)			as single
		id					as string
		pos(1 to 2)		as nmathf.vec3f
		pindex(1 to 2)	as uinteger
	end type
	
	type morph_struct
		vec(any)			as nmathf.vec3f
		index(any)		as uinteger 
		max_vec			as uinteger
		id					as string
	end type
	
	type model_struct
		file_name		as string
		max_vertices	as uinteger
		max_triangles	as uinteger
		max_surfaces	as uinteger
		vertices(any)	as nmathf.vec3f
		morphs(any)	   as morph_struct
		max_morphs		as uinteger
		tvertices(any)	as nmathf.vec3f
		vnormals(any)	as nmathf.vec3f
		tangent(any)	as nmathf.vec3f
		bitangent(any)	as nmathf.vec3f
		triangles(any)	as triangle_struct
		surfaces(any)	as surface_struct
		volume(any)		as nmathf.vec3f
		Volume_CapS(any)	as nmathf.vec4f
		fCnt				as integer
		Volume_CapF(any)     as nmathf.vec4f
		Cap_fCnt        as integer
		Volume_CapR(any)     as nmathf.vec4f
		Cap_rCnt        as integer
		max_bones		as uinteger
		bone(any)		as bone_struct
		pivot				as nmathf.vec3f
		dlist 			as gluint
		name 				as string
		shadow_caster	as integer
		static_shadows as integer
		volume_created	as integer
	end type

	type group_struct
		name	as string
		group	as integer
		id1		as integer
		id2		as integer
		id3		as integer
		id4		as integer
		bump_id as integer
	end type




	declare function lwobject_2model3d( byval object as lwObject ptr, texture() as nimage.texture_struct, byref force_clamp as integer ) as model_struct ptr
	declare sub material_sort( byref model as model_struct ptr )
	declare function material_sort_callback cdecl ( byval elm1 as any ptr, byval elm2 as any ptr ) as integer
	declare sub render_model( byref model as model_struct ptr, Texture() as nimage.Texture_Struct )
	declare sub render_model_tex_only( byref Model as Model_struct ptr, texture() as nimage.texture_struct )
	declare sub render_model_bone_tex_only( byref Model as Model_struct ptr, texture() as nimage.texture_struct )
	declare sub render_model_barebones( byref model as model_struct ptr )
	declare sub render_model_surface( byref model as model_struct ptr, Texture() as nimage.Texture_Struct, s as integer )
	declare sub unload_model3d( byref model as model_struct ptr )
	declare function load_model3d( byref filename as string, texture() as nimage.texture_struct, byref force_clamp as integer ) as model_struct ptr


	declare sub center_model( model as model_struct ptr )
	declare sub Calc_Normals( model as model_struct ptr )
	declare sub Calc_vNormals( model as model_struct ptr )
	declare sub Calc_Planes( model as model_struct ptr )
	declare sub Set_Poly_Neighbors( model as model_struct ptr )
	
	'functions for redering shadow volumes with end caps... aka the carmack method
	declare sub do_shadow_pass_wCaps( model as model_struct ptr, byref lp as nmathf.vec4f )
	declare sub calc_shadow_volume_wCaps ( model as model_struct ptr, byref lp as nmathf.vec4f )
	declare sub render_quick_volume_wCaps( model as model_struct ptr )
	
	'functions for rendering shadpw volumes without end caps... faster but camera cannot enter the volume
	declare sub Do_Shadow_Pass( Model as Model_struct ptr, byval lp as nmathf.Vec4f )
	declare sub Render_Quick_Volume( Model as Model_struct ptr )
	declare sub Calc_Shadow_Volume( Model as Model_struct ptr, byref lp as nmathf.vec4f )
	declare sub Begin_Shadow_Pass()
	
	
end namespace
