#include once "Newton.bi"
'#include once "fbmld.bi"
#include once "model.bi"
#include once "mathfuncf.bi"
#include once "hash.bi"

namespace nphysics

	const as single GRAVITY = 9.8

	const as integer OBJECT_RIGID_BODY = 1, OBJECT_COLLISION_TREE = 2

	type joint_struct
		Joint_Type      as integer
		nJoint          as NewtonJoint ptr
		DOF             as integer
		Body            as NewtonBody ptr
		Parent_Body     as NewtonBody ptr
		Mass            as single
		localMatrix0    as nmathf.matrix
		localMatrix1    as nmathf.matrix
		mMatrix         as nmathf.matrix
		Mod_Matrix      as nmathf.matrix
		Collision       as NewtonCollision ptr
		Pin             as nmathf.vec3f
		Pivot           as nmathf.vec3f
		Omega           as nmathf.vec3f
		Max_Omega       as nmathf.vec3f
		Collision_State as integer
		Do_Destruction  as integer
		Destroyed       as integer
		Min_Friction    as single
		Max_Friction    as single
		Angle           as single
		has_limits      as integer
		Min_Angle       as single
		Max_Angle       as single
		Model_ID        as uinteger
		ID              as uinteger
		Break_Force     as uinteger
	end type


	type fake_joint_struct
		
		pbody				as NewtonBody ptr
		poffsetmatrix	as nmathf.matrix
		cbody				as NewtonBody ptr
		coffsetmatrix	as nmathf.matrix
		special1			as integer
		special2			as integer
		mMatrix 			as nmathf.matrix
		Model_Id			as integer
		pcent				as nmathf.vec3f
		align1			as nmathf.vec3f
		align2			as nmathf.vec3f
		alignup			as nmathf.vec3f
		
	end type
	

	type transmission_struct
		
		RPM							as single
		max_gears					as integer
		max_rpm						as integer
		Torque_Curve				as single ptr
		Gear_Ratio					as single ptr
		Gear							as integer
		Differential_Ratio		as single
		Transmission_Efficiency	as single
		reverse						as integer
		
	end type
	
	
	type Tire_Struct
		local_position 	as nmathf.vec3f
		mMatrix				as nmathf.matrix
		mass					as single
		radius				as single
		model_id				as integer
		width					as single
		index 				as integer
		latgrip				as single
		latgripco			as single
		longgrip				as single
		longgripco			as single
	end type
	

	type vehicle_struct
		
		tire 						as tire_struct ptr
		max_tires				as integer
		
		transmission			as transmission_struct
		
		brake						as integer
		throttle					as single
		wind_resistance		as single
		center_of_mass 		as nmathf.vec3f
		
		rear_tire_omegaMag	as single
		suspension_length		as single
		steer_angle				as single
		steer_speed				as single
		steer_spring			as single
		front_tire				as integer
		rear_tire				as integer
		
	end type
	
	
	type Object_Struct
		
		tModel					as nmodel.Model_struct ptr
		radius              	as single
		id                  	as integer
		body                	as NewtonBody ptr
		material            	as integer
		mass                	as single
		collision           	as NewtonCollision ptr
				
		tMatrix					as nmathf.matrix
		bMatrix					as nmathf.matrix
		vec             		as nmathf.vmat
		
		cTarg						as nmathf.vec3f
		
		angle               	as nmathf.vec3f
		velocity            	as nmathf.vec3f
		
		external_torque     	as nmathf.vec3f
		
		omega						as nmathf.vec3f
		external_omega      	as nmathf.vec3f
		
		force						as nmathf.vec3f
		external_force      	as nmathf.vec3f
		
		contact_surf			as integer
		
		'designation         	as string
		max_sounds          	as integer
		sound               	as uinteger ptr
		
		morf_delta			as single
		morf					as single
		morf_targ			as integer
		cur_action			as integer
		on_ground			as integer
		contact_string		as string
		in_water				as integer
		foul					as integer
		foul_pos				as nmathf.vec3f
		last_good_pos		as nmathf.vec3f
		hit_goal				as integer
		win					as integer
		goal_time			as double
		win_time				as double
		reset_velocities	as integer
		shots					as integer
		
		declare sub rb_SetMatrix( byref tMatrix as nmathf.matrix )
		
		
	end type
	
	
	type entity_struct
		
		object					as object_struct ptr
		max_objects				as integer
		
		joint               	as joint_struct ptr
		max_joints          	as integer
		
		max_fake_joints     	as integer
		fake_joint			  	as fake_joint_struct ptr
		
		vehicle					as vehicle_struct ptr
		max_vehicles			as integer
		
		declare sub add_collision_tree( byref nWorld as newtonWorld ptr, byref model as nModel.model_struct ptr )
		declare sub add_convex_hull( byref nWorld as NewtonWorld ptr, byref model as nModel.model_struct ptr, byval mass as single )
		declare sub add_character( byref nWorld as NewtonWorld ptr, byval mass as single, byref eSize as nmathf.vec3f )
		declare sub add_camera( byref nWorld as NewtonWorld ptr, byref eSize as nmathf.vec3f )
		
		
		declare sub TreeBegin( byref nWorld as newtonWorld ptr )
		declare sub TreeEnd( byref nWorld as newtonWorld ptr )
		
	end type
	

	type Ray_Struct
		Param               as single
		Intersection_Point  as nmathf.vec3f
		Vec                 as nmathf.vmat
		Normal              as nmathf.vec3f
		Newton_ID           as NewtonBody ptr
		Object_ID           as integer
		Collision_ID        as integer
		Designation         as string
	end type


	declare sub build_collision_tree( byval nWorld as NewtonWorld ptr, byref Model as nmodel.Model_struct ptr, byref Object as Object_Struct )
	declare sub build_convex_hull( byval nWorld as NewtonWorld ptr, byref Model as nmodel.Model_struct ptr, byref Object as Object_Struct, byval Mass as dFLoat )
	declare sub build_camera( byval nWorld as NewtonWorld ptr, byref object as object_struct, byref eSize as nmathf.vec3f )
	declare sub build_character( byval nWorld as NewtonWorld ptr, byref Char as object_struct, byval Mass as single, byref eSize as nmathf.vec3f )
	declare sub character_controller cdecl( byval RBody as NewtonBody ptr )
	declare sub generic_physics_transform cdecl( byval RBody as NewtonBody ptr, byval mMatrix as dFloat ptr )
	declare sub character_physics_transform cdecl( byval RBody as NewtonBody ptr, byval mMatrix as dFloat ptr )
	declare sub generic_force_torque cdecl( byval RBody as NewtonBody ptr )
	
	declare function Character_World_Contact_Begin Cdecl( ByVal Material As NewtonMaterial Ptr, ByVal rBody1 As NewtonBody Ptr, ByVal rBody2 As NewtonBody Ptr ) As Integer
	declare function Character_World_Contact_Process Cdecl( byVal Material As NewtonMaterial Ptr, ByVal Contact As NewtonContact Ptr ) As Integer
	declare sub character_World_Contact_End Cdecl(ByVal Material As NewtonMaterial Ptr)
	
	declare sub camera_controller cdecl( byval RBody as NewtonBody ptr )
	
	declare sub rolling_friction_callback cdecl( byval rjoint as NewtonJoint ptr )
	
end namespace
