#include once "display.bi"

extern Display as ndisplay.Display_Struct

namespace ndisplay


	sub render_target_struct.lock()

		if glBindFramebufferEXT then

			glBindFramebufferEXT ( GL_FRAMEBUFFER_EXT, this.fBuffer )

			glViewport( 0, 0, this.Width, this.Height )

		end if

	end sub


	sub render_target_struct.unlock()

		if glBindFramebufferEXT then

			glBindFramebufferEXT ( GL_FRAMEBUFFER_EXT, 0 )

			glViewport( 0, 0, display.W, display.H )

		end if

	end sub


	function render_target_struct.create( byref w as uinteger, byref h as uinteger, byref has_depth as integer, byref has_stencil as integer ) as integer

		dim as uinteger status = 0

		if glBindFramebufferEXT then

			with this
				'create a texture to attach to the framebuffer
				glGenTextures (1, @.texture)
				glBindTexture (GL_TEXTURE_2D, .texture)
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
				glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, W, H, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0)
				'glGenerateMipmapEXT(GL_TEXTURE_2D)

				'create framebuffer object and attach texture
				glGenFramebuffersEXT (1, @.fBuffer)
				glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, .fBuffer)

				glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, .texture, 0)
				glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_RENDERBUFFER_EXT, .texture)

				if has_depth <> 0 then

					glGenRenderbuffersEXT(1, @.dBuffer)
					glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, .dBuffer)
					glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, w, h)
					glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, .dBuffer)

					'after a bit of research, it seems this next bit of code wont even work without gl 3.0 headers/support.
					'elseif has_stencil <> 0 then
					'GL_UNSIGNED_INT_24_8_EXT 'this is the correct extension, seems our header is becoming dated.
					'GL_UNSIGNED_INT_24_8_NV 'our header only supplies this, and i don't feel like swigging it.

					'glGenRenderbuffersEXT( 1, @.dBuffer )
					'glBindRenderbufferEXT( GL_RENDERBUFFER_EXT, .dBuffer )
					'glRenderbufferStorageEXT( GL_RENDERBUFFER_EXT, GL_UNSIGNED_INT_24_8_NV, w, h )

					'glFramebufferRenderbufferEXT( GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, .dBuffer )
					'glFramebufferRenderbufferEXT( GL_FRAMEBUFFER_EXT, GL_STENCIL_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, .dBuffer )

				end if


				status = glCheckFramebufferStatusEXT( GL_FRAMEBUFFER_EXT )

				if status <> GL_FRAMEBUFFER_COMPLETE_EXT then
					return 0
				end if

				this.width  = W
				this.height = H
				glBindFramebufferEXT ( GL_FRAMEBUFFER_EXT, 0 )
				return 1
			end with

		else
			return 0
		end if

	end function


	sub render_target_struct.destroy()

		with this

			if .fBuffer then
				glDeleteFramebuffersEXT( 1, @.fBuffer)
				glDeleteTextures( 1, @.texture )
			end if

			if .dBuffer then
				glDeleteRenderbuffersEXT(1, @.dBuffer)
			end if

		end with

	end sub


	sub set_ortho( byval W as integer, byval H as integer )
		glmatrixmode( Gl_PROJECTION )
		glpushmatrix()
		glloadidentity()
		glmatrixmode( Gl_MODELVIEW )
		glpushmatrix()
		glloadidentity()
		glortho( 0, W, 0, H, -1, 1 )
	end sub


	sub drop_ortho()
		glmatrixmode( Gl_PROJECTION )
		glpopmatrix()
		glmatrixmode( Gl_MODELVIEW )
		glpopmatrix()
	end sub


	sub init_light( byref index as integer, byref light as Light_Struct, byref ambient as nimage.RGBALpha, byref diffuse as nimage.RGBAlpha, byref specular as nimage.RGBAlpha, byref position as nmathf.vec4f, byref direction as nmathf.vec3f )

		light.ambient = ambient
		light.diffuse = diffuse
		light.specular = specular
		light.position = position
		light.direction = direction

		glEnable( GL_LIGHT0 + index )
		glEnable( GL_LIGHTING )

		'slightly better specular reflections without using a shader...
		glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR)

		gllightfv( GL_LIGHT0 + index, GL_AMBIENT, @light.ambient.r )
		gllightfv( GL_LIGHT0 + index, GL_DIFFUSE, @light.diffuse.r )
		gllightfv( GL_LIGHT0 + index, GL_SPECULAR, @light.specular.r )
		gllightfv( GL_LIGHT0 + index, GL_POSITION, @light.position.x )

	end sub


	sub make_font_dlist( byref DList as GlUINT,  byval Wtiles as single, byval Htiles as single, byval t_Scale as integer )
		DList = glgenlists(256)
		dim as integer i
		dim as single Tx, Ty, XStepp = 1/Wtiles, YStepp = 1/Htiles
		glMatrixMode( GL_MODELVIEW )
		glLoadidentity()
		glPushmatrix()
		i=0
		for TY as integer = Htiles-1 to 0 step -1
			for TX as integer = 0 to Wtiles-1
				glNewList DList+i, GL_COMPILE
				glBegin( GL_QUADS )

				glTexCoord2d( Tx*XStepp, (Ty+1)*YStepp )
				glVertex2i( 0, t_Scale )

				glTexCoord2d( Tx*XStepp, Ty*YStepp )
				glVertex2i( 0, 0 )

				glTexCoord2d( (Tx+1)*XStepp, Ty*YStepp )
				glVertex2i( t_Scale, 0 )

				glTexCoord2d( (Tx+1)*XStepp, (Ty+1)*YStepp )
				glVertex2i( t_Scale, t_Scale )

				glEnd()

				glTranslatef( t_Scale, 0, 0 )
				glEndList()
				i+=1
			next
		next
		glPopMatrix()
	end sub


	sub glprint( byref Strng as string, byval Texture as Gluint, byval Text_List as GLUINT, byval X as single, byval Y as single, byval R as single, byval G as single, byval B as single )

		dim as integer i, Length = len(Strng)
		'glEnable( GL_TEXTURE_2D )
		'glBindTexture( GL_TEXTURE_2D, Texture )
		'glEnable( GL_BLEND )
		'glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA )
		'glPolygonMode( GL_FRONT, GL_FILL )
		'glMatrixMode( GL_PROJECTION )
		'glPushMatrix()
		'glLoadIdentity()
		'glMatrixMode( GL_MODELVIEW )
		'glPushMatrix()
		'glLoadIdentity()
		'glOrtho( 0,Display.W, 0, Display.H, -1, 1 )
		'glDisable( GL_LIGHTING )
		'glDisable( GL_DEPTH_TEST )

		glTranslatef( X, Y, 0 )
		glColor3f( R,G,B )
		glListbase( Text_List )
		glCallLists( Length, GL_UNSIGNED_BYTE, @Strng )

		'glMatrixMode( GL_PROJECTION )
		'glPopMatrix()
		'glmatrixmode( GL_MODELVIEW )
		'glPopMatrix()
		'glenable( GL_LIGHTING )
		'glEnable( GL_DEPTH_TEST )
		'glDisable( GL_BLEND )

	end sub


	function init_gl_window( byval W as integer, byval H as integer, byval BPP as integer, byval Num_buffers as integer, byval Num_Samples as integer, byval Fullscreen as integer ) as integer

		dim Flags as integer = FB.GFX_OPENGL

		if FullScreen then
			Flags or = FB.GFX_FULLSCREEN
		end if

		screencontrol FB.SET_GL_COLOR_BITS, BPP
		screencontrol FB.SET_GL_DEPTH_BITS, 24
		screencontrol FB.SET_GL_STENCIL_BITS, 8

		if Num_Samples>0 then
			screencontrol FB.SET_GL_NUM_SAMPLES, Num_Samples
			Flags or = FB.GFX_MULTISAMPLE
		end if

		screenres W, H, BPP, Num_Buffers, Flags or FB.GFX_STENCIL_BUFFER

'		Display.glVer = glGetString( GL_VERSION )

		glviewport( 0, 0, W, H )
		Display.W = W
		Display.H = H
		Display.W2 = W\2
		Display.H2 = H\2
		Display.BPP = BPP
		Display.FOV = 60
		Display.Aspect = W/H
		Display.znear = 1
		Display.zfar = 1000

		display.pMatrix.InfiniteProjection( display.fov, display.aspect, display.znear )
		'display.pMatrix.perspective( display.fov, display.aspect, display.znear, display.zFar )
		glMatrixMode( GL_PROJECTION )
		glLoadIdentity()
		glLoadMatrixf( display.pMatrix )

		glClearColor( .52734375, .8046875, .9765625, 0.5 )
		glClearDepth( 1.0 )
		glEnable( GL_DEPTH_TEST )
		glDepthFunc( GL_LEQUAL )
		glEnable( GL_COLOR_MATERIAL )

		glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST )
		glPolygonmode( GL_FRONT_AND_BACK, GL_FILL )
		glEnable( GL_CULL_FACE )
		glMatrixMode( GL_MODELVIEW )
		glShadeModel( GL_SMOOTH )
		gather_extensions()
		
		build_stencil_cube( display.stencil_cube )
		
		return 1

	end function


	sub gather_extensions()

		dim extensions as string
		screencontrol FB.GET_GL_EXTENSIONS, extensions

		dim as integer i, j

		do
			i+=1
			j=i
			dim as string tstring
			dim as string char
			do
				char = mid(extensions,j,1)
				tstring+=char
				j=j+1
				if char = " " then
					exit do
				end if
			loop while j<len(extensions)

			i = j-1
			'print #50, tstring

		loop while i<len(extensions)


		if (instr(extensions, "GL_EXT_texture_filter_anisotropic") <> 0) then
			_anisotropic_ = 1
		end if

		if (instr(extensions, "GL_EXT_framebuffer_object") <> 0) then
			print #50, "GL_EXT_framebuffer_object is supported"
			_framebuffer_ = 1
			glGenFramebuffersEXT            = screenglproc("glGenFramebuffersEXT")
			glDeleteFramebuffersEXT         = screenglproc("glDeleteFramebuffersEXT")
			glBindFramebufferEXT            = screenglproc("glBindFramebufferEXT")
			glFramebufferTexture2DEXT       = screenglproc("glFramebufferTexture2DEXT")
			glFramebufferRenderbufferEXT    = screenglproc("glFramebufferRenderbufferEXT")
			glGenRenderbuffersEXT           = screenglproc("glGenRenderbuffersEXT")
			glBindRenderbufferEXT           = screenglproc("glBindRenderbufferEXT")
			glRenderbufferStorageEXT        = screenglproc("glRenderbufferStorageEXT")
			glDeleteRenderbuffersEXT		  = screenglproc("glDeleteRenderbuffersEXT")
			glCheckFramebufferStatusEXT	  = screenglproc("glCheckFramebufferStatusEXT")
		else
			print #50, "GL_EXT_framebuffer_object is NOT supported"
		end if


		if (instr(extensions, "GL_ARB_multitexture") <> 0) then
			print #50, "GL_ARB_multitexture is supported"
			_multitexture_ = 1
			glMultiTexCoord2fARB            = screenglproc("glMultiTexCoord2fARB")
			glMultiTexCoord2fvARB           = screenglproc("glMultiTexCoord2fvARB")
			glActiveTextureARB              = screenglproc("glActiveTextureARB")
			glClientActiveTextureARB        = screenglproc("glClientActiveTextureARB")
			glgetintegerv( GL_MAX_TEXTURE_UNITS_ARB, @maxTexelUnits )
			print #50, maxTexelUnits & " Texture units supported"
			if maxTexelUnits<3 then
				_multitexture_ = 0
				print #50, "Insufficient texture units for multitexture support"
			end if

		else
			print #50, "GL_ARB_multitexture is NOT supported"
		end if


		if (instr(extensions, "GL_ARB_shading_language_100") <> 0) then
			print #50, "GL_ARB_shading_language_100 is supported"
			if (instr(extensions, "GL_ARB_vertex_program") <> 0) then
				print #50, "GL_ARB_vertex_program is supported"
				if (instr(extensions, "GL_ARB_fragment_program") <> 0) then
					print #50, "GL_ARB_fragment_program is supported"
					_shader100_ = 1
					glCreateShaderObjectARB     = screenglproc("glCreateShaderObjectARB")
					glShaderSourceARB           = screenglproc("glShaderSourceARB")
					glGetShaderSourceARB        = screenglproc("glGetShaderSourceARB")
					glCompileShaderARB          = screenglproc("glCompileShaderARB")
					glDeleteObjectARB           = screenglproc("glDeleteObjectARB")
					glCreateProgramObjectARB    = screenglproc("glCreateProgramObjectARB")
					glAttachObjectARB           = screenglproc("glAttachObjectARB")
					glUseProgramObjectARB       = screenglproc("glUseProgramObjectARB")
					glLinkProgramARB            = screenglproc("glLinkProgramARB")
					glValidateProgramARB        = screenglproc("glValidateProgramARB")
					glGetInfoLogARB             = screenglproc("glGetInfoLogARB")
					glGetObjectParameterivARB   = screenglproc("glGetObjectParameterivARB")
					glGetUniformLocationARB     = screenglproc("glGetUniformLocationARB")
					glUniform1iARB              = screenglproc("glUniform1iARB")
					glUniform1fARB              = screenglproc("glUniform1fARB")
					glUniform2fvARB             = screenglproc("glUniform2fvARB")
					glUniform3fvARB             = screenglproc("glUniform3fvARB")
				else
					print #50, "GL_ARB_fragment_program is NOT supported"
				end if
			else
				print #50, "GL_ARB_vertex_program is NOT supported"
			end if
		else
			print #50, "GL_ARB_shading_language_100 is NOT supported"
		end if
		
		glActiveStencilFaceEXT = screenglproc("glActiveStencilFaceEXT")
		if glActiveStencilFaceEXT = 0 then
			_stencil2_ = 0
			print #50, "glActiveStencilFaceEXT is NOT supported"
		else
			print #50, "glActiveStencilFaceEXT is supported"
			_stencil2_ = 1
		end if
		
	end sub


	function init_shader( File_Name as string, Shader_Type as integer )as GlHandleARB

		dim as integer i
		dim as integer Line_Cnt
		dim as string Shader_Text, tString
		dim as Gluint Shader_Compile_Success
		dim as GlHandleARB Shader
		dim as uinteger FileNum = freefile

		open File_Name for binary as #FileNum
		do while not eof(FileNum)
			line input #FileNum, tString
			Shader_Text += tString + chr( 13, 10 )
		loop
		close #FileNum

		dim as GLcharARB ptr table(0) => { strptr( Shader_Text ) }
		Shader = glCreateShaderObjectARB( Shader_Type )
		glShaderSourceARB( Shader, 1, @table(0), 0 )
		glCompileShaderARB( Shader )

		glGetObjectParameterivARB( Shader, GL_OBJECT_COMPILE_STATUS_ARB, @Shader_Compile_Success )
		if Shader_Compile_Success = 0 then
			dim as Gluint infologsize
			glGetObjectParameterivARB( Shader, GL_OBJECT_INFO_LOG_LENGTH_ARB, @infoLogSize)
			dim as GlByte infolog(InfoLogSize)
			glGetInfoLogARB( Shader, InfoLogSize, 0, @infoLog(0))
			tString=""
			for i = 0 to InfoLogSize-1
				tString+=chr(InfoLog(i))
			next
			print #50, "Shader Infolog error message:"
			print #50, tString
			return 0
		else
			return Shader
		end if

	end function

	sub build_stencil_cube( byref List as gluint )
		List = glGenLists(1)
		glNewList( List, GL_COMPILE )
		'glFrontFace(GL_CCW)
		glColorMask(1, 1, 1, 1)
		glColor4f(0.0, 0.0, 0.0, 0.5)
		glEnable(GL_BLEND)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
		glStencilFunc( GL_NOTEQUAL, 0, &hffffffff )
		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP)
		'glpushmatrix()
		glloadidentity()
		glbegin(GL_TRIANGLE_STRIP)
		glVertex3f(-1, 1, -1)
		glVertex3f(-1,-1, -1)
		glVertex3f( 1, 1, -1)
		glVertex3f( 1,-1, -1)
		glend()
		'glpopmatrix()

		glDisable(GL_BLEND)
		glDepthFunc(GL_LEQUAL)
		glDepthMask(GL_TRUE)
		glEnable(GL_LIGHTING)
		glDisable(GL_STENCIL_TEST)
		glShadeModel(GL_SMOOTH)
		glEndList()
	end sub


	constructor font()
		
		this.scale = 1
		
		glgentextures( 1, @this.texture )
	
		this.dList = glGenLists(256)
		
			
		dim as integer i = 0
		dim as single Tx, Ty, XStepp = 1/16, YStepp = 1/16
		glMatrixMode( GL_MODELVIEW )
		glLoadidentity()
		glPushmatrix()
		
		for TY as integer = 15 to 0 step -1
			for TX as integer = 0 to 15
				glNewList( this.dList+i, GL_COMPILE )
				glBegin( GL_QUADS )

				glTexCoord2d( Tx*XStepp, (Ty+1)*YStepp )
				glVertex2f( 0, this.scale )

				glTexCoord2d( Tx*XStepp, Ty*YStepp )
				glVertex2f( 0, 0 )

				glTexCoord2d( (Tx+1)*XStepp, Ty*YStepp )
				glVertex2f( this.scale, 0 )

				glTexCoord2d( (Tx+1)*XStepp, (Ty+1)*YStepp )
				glVertex2f( this.scale, this.scale )

				glEnd()

				glTranslatef( this.scale/1.5, 0, 0 )
				glEndList()
				i+=1
			next
		next
		glPopMatrix()
		
	end constructor
	
	destructor font()
		glDeleteLists( this.dList, 1 )
		glDeleteTextures( 1, @this.texture )
	end destructor
	
	
	sub font.oBegin()
		glEnable( GL_TEXTURE_2D )
		glBindTexture( GL_TEXTURE_2D, this.texture )
		glEnable( GL_BLEND )
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA )
		glPolygonMode( GL_FRONT, GL_FILL )
		glMatrixMode( GL_PROJECTION )
		glPushMatrix()
		glLoadIdentity()
		glMatrixMode( GL_MODELVIEW )
		glPushMatrix()
		glLoadIdentity()
		glOrtho( 0,Display.W, 0, Display.H, -1, 1 )
		glDisable( GL_LIGHTING )
		glDisable( GL_DEPTH_TEST )
	end sub
	
	sub font.oEnd()
		glMatrixMode( GL_PROJECTION )
		glPopMatrix()
		glmatrixmode( GL_MODELVIEW )
		glPopMatrix()
		glenable( GL_LIGHTING )
		glEnable( GL_DEPTH_TEST )
		glDisable( GL_BLEND )
	end sub
	
	sub font.oPrint( Byval Strng As String, Byval X As Single, Byval Y As Single, Byval R As single = 1.0, Byval G As single = 1.0, Byval B As single = 1.0, byval s as single = 1.0 )
		
		if s = 0 then s = this.scale
		
		glPushMatrix()
		glTranslatef( X, Y, 0 )
		glScalef(s,s,1)
		glColor3f( R,G,B )
		glListbase( this.dList )
		glCallLists( len(strng), GL_UNSIGNED_BYTE, strptr(Strng) )
		glPopMatrix()
		
	end sub
	
	sub font.load( byref fileName as string )
		
		dim as string tstring
		dim as integer lastslash

		for i as integer = 1 to len(filename)
			if mid(filename,i,1) = "/" then
				lastslash=i
			end if
		next

		tstring = mid(filename, lastslash + 1 )
		
		glbindtexture( GL_TEXTURE_2D, this.texture )
		glenable( GL_TEXTURE_2D )

		dim SprWidth as integer
		dim SprHeight as Integer
		dim as any ptr temp

				
		if temp=0 Then
			png_dimensions( "res/texture/" & tstring, SprWidth, SprHeight )
			temp = png_load( "res/texture/" & tstring, PNG_TARGET_OPENGL )
		end if

		glteximage2d( GL_TEXTURE_2D, 0, GL_RGBA, SprWidth, SprHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, temp )

		glubuild2dmipmaps( GL_TEXTURE_2D, GL_RGBA, SprWidth, SprHeight, GL_RGBA, GL_UNSIGNED_BYTE, temp )

		'if force_clamp then
			gltexparameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE )
			gltexparameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE )
		'else
		'	gltexparameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT )
		'	gltexparameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT )
		'end if

		gltexparameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR )
		gltexparameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR )
		gltexenvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)


		'if use_anisotropic <> 0 then
		'	dim as glFloat Largest_Texture_Filter
		'	glgetfloatv( GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, @Largest_Texture_Filter )
		'	gltexparameterf( GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, Largest_Texture_Filter )
		'end if

		png_destroy( temp )
		
	end sub
	
end namespace
