'#define con_print

'#include once "fbmld.bi"
#include once "fbgfx.bi"
#include once "mathfuncf.bi"
#include once "model.bi"
#include once "image.bi"
#include once "model.bi"
#include once "objmodel.bi"
#include once "display.bi"
#include once "physics.bi"
#include once "hash.bi"

using ndisplay
using nmodel
using nimage
using nmathf
using nphysics
using nhash

extern display as display_struct
extern h_table	as ht_t ptr

dim as display_struct display
dim as ht_t ptr h_table



dim shared as double frame_steps
dim shared as double accumulative_step, frequency, newt_time, time_step
dim shared as double sim_speed

type mousetype
	as integer x,y,lx,ly,b,lb,r,lr
end type

declare function main( byval argc as integer, byval argv as zstring ptr ptr ) as integer
declare sub prep_1st_pass(byref light as light_struct)
declare sub render_1st_pass( byref model as model_struct ptr, byref tMatrix as matrix, texture() as texture_struct )
declare sub prep_2nd_pass()
declare sub render_2nd_pass( byref model as model_struct ptr, byref tMatrix as matrix, byref light as light_struct )
declare sub prep_3rd_pass( byref light as light_struct )
declare sub render_3rd_pass( byref model as model_struct ptr, byref tMatrix as matrix, texture() as texture_struct )
declare sub load_shaders()
declare Sub Create_Shadow_Buffer( Byref Texture As Gluint )


declare function find_surface_center( model as model_struct ptr, surfname as string ) as vec3f
declare sub resort_model_surfaces( model as model_struct ptr )
declare function Spline_Point( Points() as vec3f, byval p as single ) as vec3f

function load_level( fName as string, byref nWorld as newtonWorld ptr, byref world as entity_struct, byref plr as entity_struct, byref cam as entity_struct, byref model as model_struct ptr, texture() as texture_struct ) as integer

	if nWorld then
		newtonDestroy( nWorld )
	end if

	if model then
		unload_model3d( model )
	end if

	model = load_model3d( "res/model/" & fname, Texture(), 0 )
	resort_model_surfaces( model )
	model->static_shadows = true

	nworld = NewtonCreate(0,0)
	#ifdef __FB_LINUX__
	NewtonSetPlatformArchitecture(nworld, 0)
	#endif
	NewtonSetMinimumFrameRate( nworld, frame_steps )

	world.max_objects = 0
	world.add_collision_tree( nWorld, model )

	plr.max_objects = 0
	plr.add_character( nWorld, 5, vec3f(.5,.5,.5) )
	plr.object[0].win = 0


	if plr.joint then
		plr.max_joints = 0
		deallocate(plr.joint)
	end if

	plr.max_joints = 1
	plr.joint = callocate( plr.max_joints*sizeof(joint_struct))
	plr.joint[0].nJoint = NewtonConstraintCreateUserJoint( nWorld, 1, @Rolling_Friction_Callback(), plr.object[0].body, 0 )
	NewtonJointSetUserData( plr.joint[0].nJoint, @plr.object[0] )
	var angdamp = vec3f(.05,.05,.05)
	NewtonBodySetAngularDamping( plr.object[0].body, @angdamp.x )
	NewtonBodySetLinearDamping( plr.object[0].body, .05 )
	NewtonBodyCoriolisForcesMode( plr.object[0].body, 1 )

	cam.max_objects = 0
	cam.add_camera( nWorld, vec3f(1,1,1) )
	'add_camera( nWorld, cam, vec3f(3f,3f,3f) )

	plr.object[0].tMatrix.Position = vec3f( 0, 26f, 0 )
	'plr.object[0].tMatrix.Rotate( 0, 95, 0 )

	'newtonBodySetMatrix( plr.object[0].Body, plr.object[0].tMatrix )
	plr.object[0].rb_SetMatrix( plr.object[0].tMatrix )


	'time for some materials...
	'everyting else uses default as the base for calculations, so we need to set it up first
	Dim As Integer Default_Material = NewtonMaterialGetDefaultGroupID (nWorld)
	NewtonMaterialSetDefaultSoftness (nWorld, Default_Material, Default_Material, .75)
	NewtonMaterialSetDefaultElasticity (nWorld, Default_Material, Default_Material, .5)
	NewtonMaterialSetDefaultCollidable (nWorld, Default_Material, Default_Material, 1)
	NewtonMaterialSetDefaultFriction (nWorld, Default_Material, Default_Material, 1, .9)

	world.object[0].material = NewtonMaterialCreateGroupID (nWorld)
	NewtonMaterialSetDefaultSoftness (nWorld,   world.object[0].material, Default_Material, .75)
	NewtonMaterialSetDefaultElasticity (nWorld, world.object[0].material, Default_Material, .25)
	NewtonMaterialSetDefaultCollidable (nWorld, world.object[0].material, Default_Material, 1)
	NewtonMaterialSetDefaultFriction (nWorld,   world.object[0].material, Default_Material, 1, .9)
	NewtonBodySetMaterialGroupID( world.object[0].Body, world.object[0].material )


	plr.object[0].material = NewtonMaterialCreateGroupID (nWorld)
	NewtonMaterialSetDefaultSoftness (nWorld,   plr.object[0].material, world.object[0].material, .75)
	NewtonMaterialSetDefaultElasticity (nWorld, plr.object[0].material, world.object[0].material, .25)
	NewtonMaterialSetDefaultCollidable (nWorld, plr.object[0].material, world.object[0].material, 1)
	NewtonMaterialSetDefaultFriction (nWorld,   plr.object[0].material, world.object[0].material, 1, .9)
	NewtonMaterialSetCollisionCallback (nWorld, plr.object[0].material, world.object[0].material, @plr.object[0], @Character_World_Contact_Begin, @Character_World_Contact_Process, @character_World_Contact_End )
	NewtonBodySetMaterialGroupID( plr.object[0].Body, plr.object[0].material )
	NewtonMaterialSetContinuousCollisionMode( nWorld, plr.object[0].material, world.object[0].material, 1 )


	cam.object[0].Material = NewtonMaterialCreateGroupID (nWorld)
	NewtonMaterialSetDefaultSoftness (nWorld,   cam.object[0].material, world.object[0].material, .0000001)
	NewtonMaterialSetDefaultElasticity (nWorld, cam.object[0].material, world.object[0].material, 0)
	NewtonMaterialSetDefaultCollidable (nWorld, cam.object[0].material, world.object[0].material, 1)
	NewtonMaterialSetDefaultCollidable (nWorld, cam.object[0].material, Default_Material, 0)
	NewtonMaterialSetDefaultFriction (nWorld,   cam.object[0].material, world.object[0].material, .00001, .00001)
	NewtonMaterialSetContinuousCollisionMode( nWorld, cam.object[0].material, world.object[0].material, 1 )

	NewtonBodySetMaterialGroupID( cam.object[0].Body, cam.object[0].material )
	NewtonMaterialSetDefaultCollidable( nWorld, cam.object[0].material, plr.object[0].material, 0 )


	plr.object[0].tMatrix.loadIdentity()
	dim as vec3f tpos = find_surface_center( model, "tee box" )
	tpos.y+=1
	plr.object[0].tMatrix.Position = tpos
	dim as vec3f gpos = find_surface_center( model, "goal" )
	dim as single tang = atan2(gpos.x-tpos.x, gpos.z-tpos.z)*inv_pi_180
	plr.object[0].tMatrix.Rotate(0f,tang,0f)
	NewtonBodySetMatrix( plr.object[0].Body, plr.object[0].tMatrix )
	plr.object[0].bMatrix.Rotate(0f,tang,0f)
	plr.object[0].last_good_pos = tpos
	cam.object[0].tMatrix.Position = plr.object[0].bMatrix.Position+plr.object[0].bMatrix.up*15f
	NewtonBodySetMatrix( cam.object[0].Body, cam.object[0].tMatrix )
	return 1

end function

end main( __FB_ARGC__, __FB_ARGV__ )


function main( byval argc as integer, byval argv as zstring ptr ptr ) as integer

	use_shaders = true
	use_multitex = true
	use_anisotropic = true

	open cons for output as #50

	h_table = new ht_t( 256 )

	dim as texture_struct texture(any)
	'redim texture(0)

	frame_steps = 300
	frequency = 1d/frame_steps
	sim_speed = 2.5
	dim as double hazard_timer

	dim as vmat vCam
	dim as model_struct ptr tmodel(4)
	dim as mousetype		m
	dim as integer 		cam_mode = 0
	dim as integer 		clickX, clickY
	dim as single			zoom = 10, rx, ry, py, px
	dim as light_struct light
	dim as newtonWorld ptr nWorld
	dim as entity_struct world, cam, plr
	dim as double shot_time
	'dim as entity_struct crates(50)

	init_gl_window( 640, 480, 32, 0, 4, 0 )

	dim as font myFont
	myFont.Load( "res/texture/text2.png" )


	init_light( 0, light, type(.35,.35,.35,1.0), type(.5,.5,.5,1.0), type(.85,.85,.85,1.0), vec4f(20f,150f,20f,1f), vec3f(0f,1f,0f) )
	glEnable( GL_VERTEX_ARRAY ) 'enabled for the duration of the game... the less state switching the better
	load_shaders()
	load_texture( "res/texture/aim.png", texture(), 0  )
	dim as integer texAim = ubound(texture)
	load_texture( "res/texture/power.png", texture(), 0  )
	dim as integer texPower = ubound(texture)
	load_texture( "res/texture/power_cursor.png", texture(), 0  )
	dim as integer texPowerCursor = ubound(texture)
	load_texture( "res/texture/target.png", texture(), 0  )
	dim as integer texTarget = ubound(texture)

	dim as single ratio = (display.w/500)
	

	chdir exepath

	tmodel(1) = load_model3d( "res/model/wesley.lwo", Texture(), 0 )
	tmodel(2) = load_model3d( "res/model/wesley_ball.lwo", Texture(), 0 )
	tmodel(3) = load_model3d( "res/model/arrowhead.lwo", Texture(), 0 )
	tmodel(4) = load_model3d( "res/model/skybox.lwo", Texture(), 0 )

	load_level( "course1.lwo", nWorld, world, plr, cam, tModel(0), texture() )
	vcam.p = cam.object[0].tMatrix.Position
	
	newt_time = timer
	'test the morph on caveman
	plr.object[0].morf_delta = 50
	'plr.object[0].morf_targ = 0
	dim as single action_var
	dim as integer acting

	dim as single vertical, horizontal, arrowlen, lookAhead
	dim as single impulse_strength, impulse_delta = 35
	dim as vec3f spline(3)
	dim as vec3f omega
	dim as vec3f pointdeltavec
	dim as vec3f pointposit
	dim as single tdist
	dim as integer swing_stage = 0, space_down, hazard_reset
	dim as string print_foul

	dim as integer do_shot
	dim as integer fly_through = true, fly_toggle, fly_count
	dim as vec3f tpos = find_surface_center( tModel(0), "tee box" )
	dim as vec3f gpos = find_surface_center( tModel(0), "goal" )
	dim as vec3f flycam_Look, targ, lookTarg, camvec, backvec
	
	
	newt_time = timer
	do

		time_step = timer - newt_time
		'if not fly_through then
		cam.object[0].cTarg = plr.object[0].bMatrix.Position-plr.object[0].bMatrix.forward*zoom
		cam.object[0].cTarg.y = plr.object[0].bMatrix.Position.y+(zoom*1.5)
		'end if

		accumulative_step += time_step*sim_speed

		while accumulative_step >= frequency

			NewtonUpdate(nworld, frequency)
			accumulative_step -= frequency

		wend

		newt_time = timer

		if multikey(FB.SC_S) then
			use_shaders = not use_shaders
		end if

		if not fly_through then

			if swing_stage = 1 then

				if multikey(FB.SC_UP) then
					vertical+=1*time_step
					vertical = int(vertical*100)/100
					if vertical>1 then
						vertical = 1
					end if
				end if

				if multikey(FB.SC_DOWN) then
					vertical-=1*time_step
					vertical = int(vertical*100)/100
					if vertical<-1 then
						vertical=-1
					end if
				end if

				if multikey(FB.SC_RIGHT) then
					horizontal+=1*time_step
					horizontal = int(horizontal*100)/100
					if horizontal>1 then
						horizontal=1
					end if
				end if

				if multikey(FB.SC_LEFT) then
					horizontal-=1*time_step
					horizontal = int(horizontal*100)/100
					if horizontal<-1 then
						horizontal=-1
					end if
				end if

			elseif swing_stage = 0 then

				if multikey(FB.SC_RIGHT) then
					plr.object[0].bMatrix.rotate(0f,-75f*time_step,0f)
				end if

				if multikey(FB.SC_LEFT) then
					plr.object[0].bMatrix.rotate(0f,75f*time_step,0f)
				end if
				
				if multikey(FB.SC_UP) then
					lookAhead+=time_step*10f
					if lookAhead>10 then lookAhead = 10
				else
					lookAhead-=time_step*10f
					if lookAhead<0 then lookAhead = 0
				end if
				
			end if

			if swing_stage = 2 and do_shot = 0 then
				impulse_strength+=impulse_delta*time_step

				if impulse_strength>50 then
					impulse_strength = 50
					impulse_delta *= -1
				elseif impulse_strength<0 then
					impulse_strength = 0
					impulse_delta *= -1
				end if

			end if


			'setup the launch vector, strength and spin...
			'if do_shot = 0 then
			omega = plr.object[0].bMatrix.up*(horizontal*impulse_strength)
			omega += plr.object[0].bMatrix.forward*(-horizontal*impulse_strength)
			pointdeltavec = plr.object[0].bMatrix.forward + (plr.object[0].bMatrix.up*-vertical)
			pointdeltavec.normalize
			pointposit = plr.object[0].bMatrix.position+pointdeltavec
			tdist = pointposit.distance(plr.object[0].bMatrix.position)
			pointdeltavec*=impulse_strength
			'end if

			if plr.object[0].velocity.magnitude()<.01 and swing_stage = 3 and plr.object[0].foul = 0 and plr.object[0].hit_goal = 0 then
				swing_stage = 0
			end if

			If MultiKey(fb.SC_SPACE) then
				if not space_down then
					space_down = true

					select case swing_stage
						case 0
							swing_stage = 1
						case 1
							swing_stage = 2
						case 2
							do_shot = true
							shot_time = timer+1
					end select
				end if

			else
				space_down = false
			end if

			if do_shot then
				if timer>shot_time then
					NewtonAddBodyImpulse( plr.object[0].body, @pointdeltavec.x, @pointposit.x )
					NewtonBodySetOmega( plr.object[0].body, @omega.x )
					swing_stage = 3
					horizontal = 0
					vertical = 0
					impulse_strength = 0
					plr.object[0].shots+=1
					do_shot = false
				end if
			end if

		end if

		if plr.object[0].in_water = 1 then
			print_foul = "Water Hazard!"
			if hazard_reset = false then
				hazard_timer = timer+5
				hazard_reset = true
			else
				if timer>hazard_timer then
					hazard_reset = false
					print_foul = ""
					plr.object[0].reset_velocities = true
					plr.object[0].foul = 0
					plr.object[0].in_water = 0
					plr.object[0].tMatrix.position = plr.object[0].last_good_pos
					NewtonBodySetMatrix( plr.object[0].Body, plr.object[0].tMatrix )
					swing_stage = 0
				end if
			end if

		elseif plr.object[0].foul = 2 then
			print_foul = "Out Of Bounds"
			if hazard_reset = false then
				hazard_timer = timer+5
				hazard_reset = true
			else
				if timer>hazard_timer then
					hazard_reset = false
					print_foul = ""
					plr.object[0].reset_velocities = true
					plr.object[0].foul = 0
					plr.object[0].tMatrix.position = plr.object[0].last_good_pos
					NewtonBodySetMatrix( plr.object[0].Body, plr.object[0].tMatrix )
					swing_stage = 0
				end if
			end if
		end if

		dim as single velmag = plr.object[0].velocity.magnitude()*1.5
		if velmag<.1 then
			plr.object[0].morf_delta = 0
		end if

		plr.object[0].morf_delta+= (velmag*time_step)
		plr.object[0].morf = sin(plr.object[0].morf_delta)

		if plr.object[0].morf_delta>pi then
			plr.object[0].morf_delta = 0
			plr.object[0].morf_targ+=1
			if plr.object[0].morf_targ>1 then plr.object[0].morf_targ = 0
		end if



		m.lx = m.x : m.ly = m.y : m.lb = m.b : m.lr = m.r
		GetMouse m.x,m.y,m.r,m.b
		If m.x = -1 Then m.x=m.lx : m.y=m.ly : m.r=m.lr : m.b=m.lb


		if cam_mode = 0 then

			If m.b = 0 andAlso m.lb > 0 Then
				SetMouse ,,1
			EndIf

			zoom+=(m.r-m.lr)
			if zoom<10 then zoom = 10
			if zoom>50 then zoom = 50

		elseif cam_mode = 1 then

			setmouse( display.w2, display.h2, 0 )

			rx = -(m.x - display.w2)/1000
			ry = -(m.y - display.h2)/1000

			vCam.rotate( ry, rx, -vCam.r.y )

			if multikey( FB.SC_W ) then
				vCam.p+=vCam.f/5f
			end if

			if multikey( FB.SC_S ) then
				vCam.p-=vCam.f/5f
			end if

			if multikey( FB.SC_A ) then
				vCam.p-=vCam.r/5f
			end if

			if multikey( FB.SC_D ) then
				vCam.p+=vCam.r/5f
			end if

		end if

		'build the launch spline
		if swing_stage=2 then
			dim as single nImpulse = impulse_strength*1.5

			dim as vec3f dvec = pointdeltavec
			dvec.normalize
			spline(0) = plr.object[0].tMatrix.position
			spline(1) = plr.object[0].tMatrix.position+(dvec*(nImpulse/2))
			spline(3) = spline(1)
			spline(2) = plr.object[0].tMatrix.position+(dvec*nImpulse)
			spline(2).y = spline(0).y
		end if

		dim as matrix lMatrix

		if fly_through then

			if fly_toggle = 0 then
				backvec = gpos-tpos
				backvec.normalize
				targ = gpos+backvec*15
				targ.y+=15
				camvec = targ-vcam.p
				camvec.normalize
				vcam.p+= camvec*time_step*10

				dim as single dist = vcam.p.distance(targ)

				if dist<=10 then
					lookTarg = tpos
				else
					lookTarg = gpos
				end if

				flycam_Look+=(lookTarg-flycam_Look)*time_step

				if vcam.p.distance(targ)<=2 then
					fly_toggle+=1
					fly_count+=1
				end if

			elseif fly_toggle = 1 or fly_toggle = 3 then
				backvec = tpos.cross(gpos)
				backvec.normalize
				targ = ((tpos+gpos)/2)+backvec*25
				targ.y=((tpos.y+gpos.y)/2)+15
				camvec = targ-vcam.p
				camvec.normalize
				vcam.p+= camvec*time_step*10

				dim as single dist = vcam.p.distance(targ)

				'if dist<=5 then
				'	lookTarg = gpos
				'else
				'	lookTarg = tpos
				'end if

				flycam_Look+=(lookTarg-flycam_Look)*time_step

				if dist<=2 then
					fly_toggle+=1
					if fly_toggle > 3 then fly_toggle = 0
					fly_count+=1
				end if


			elseif fly_toggle = 2 then
				backvec = tpos-gpos
				backvec.normalize
				targ = tpos+backvec*15
				targ.y+=15
				camvec = targ-vcam.p
				camvec.normalize
				vcam.p+= camvec*time_step*10

				dim as single dist = vcam.p.distance(targ)

				if dist<=10 then
					lookTarg = gpos
				else
					lookTarg = tpos
				end if

				flycam_Look+=(lookTarg-flycam_Look)*time_step

				if dist<=2 then
					fly_toggle+=1
					fly_count+=1
				end if

			end if

			lMatrix.LookAt( vcam.p, flycam_Look, vec3f(0,1,0) )
			cam.object[0].cTarg = lMatrix.Position

			if fly_count>2 then
				fly_through = 0
			end if

		else

			if swing_stage = 2 then
				vCam.p = plr.object[0].bMatrix.Position - plr.object[0].bMatrix.right * 15f
				vCam.p.y = plr.object[0].bMatrix.Position.y + zoom
				lMatrix.LookAt( vcam.p, spline(2), vec3f(0,1,0) )

			elseif swing_stage = 3 then
				vCam.p = plr.object[0].bMatrix.Position - plr.object[0].bMatrix.right * 15f
				vCam.p.y = plr.object[0].bMatrix.Position.y + zoom*1.5
				lMatrix.LookAt( vcam.p, plr.object[0].bMatrix.Position, vec3f(0,1,0) )

			else
				vcam.p = cam.object[0].tMatrix.position
				lMatrix.LookAt( vcam.p, plr.object[0].tMatrix.position+plr.object[0].bMatrix.forward*lookAhead, vec3f(0,1,0) )

			end if

		end if

		glClear( GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT )

		glLoadMatrixf( lMatrix )
		gllightfv( GL_LIGHT0, GL_POSITION, @light.position.x )


		glDepthFunc( GL_NEVER )
		glDisable( GL_DEPTH_TEST )
		glDisable( GL_LIGHTING )

		glPushMatrix()
		glTranslatef( vcam.p.x, vcam.p.y, vcam.p.z)
		glScalef(100,100,100)
		glCallList( tModel(4)->dList )
		glPopMatrix()

		glEnable( GL_DEPTH_TEST )
		glDepthFunc( GL_LEQUAL )
		glEnable( GL_LIGHTING )



		'i'll clean all of this up... the method isn't exactly pretty, but it works perfectly
		dim as matrix idMatrix
		idMatrix.LoadIdentity()


		glEnable( GL_LIGHTING )
		glEnable( GL_LIGHT0 )
		glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR)
		glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE )
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @light.ambient.r)
		glLightfv( GL_LIGHT0, GL_AMBIENT, @light.ambient.r )
		glLightfv( GL_LIGHT0, GL_DIFFUSE, @light.diffuse.r )
		glLightfv( GL_LIGHT0, GL_SPECULAR, @light.specular.r )
		glLightfv( GL_LIGHT0, GL_POSITION, @light.position.x )
		glEnable( GL_DEPTH_TEST )



		dim as matrix aMatrix

		if swing_stage = 0 then
			render_1st_pass( tModel(1), plr.object[0].bMatrix, texture() )

		elseif swing_stage = 2 then

			for i as single = 0 to 1 step 1/11
				aMatrix.pointat( spline_point(spline(), i), spline_point(spline(), i+.1) )
				render_1st_pass( tModel(3), aMatrix, texture() )
			next

			render_1st_pass( tModel(2), plr.object[0].tMatrix, texture() )

		else
			render_1st_pass( tModel(2), plr.object[0].tMatrix, texture() )

		end if

		render_1st_pass( world.object[0].tModel, idMatrix, texture() )

		if swing_stage=2 then
			glEnable( GL_TEXTURE_2D )
			glBindTexture( GL_TEXTURE_2D, texture(texTarget).img )

			glEnable( GL_BLEND )
			glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA )
			glColor3f(1,1,1)

			glpushmatrix()
			gltranslatef(spline(2).x,spline(2).y,spline(2).z)
			glBegin(GL_QUADS)
			gltexcoord2f(0,0)
			glvertex3f(-1,0,-1)

			gltexcoord2f(0,1)
			glvertex3f(-1,0,1)

			gltexcoord2f(1,1)
			glvertex3f(1,0,1)

			gltexcoord2f(1,0)
			glvertex3f(1,0,-1)
			glEnd()
			glpopmatrix()
			glDisable( GL_TEXTURE_2D )
		end if


		'#if 1
		begin_Shadow_Pass()

		dim as matrix iMatrix = idMatrix.Inverse()
		dim as vec4f tLp = light.position*iMatrix
		glpushmatrix()
		glMultMatrixf( idMatrix )
		do_shadow_pass( world.object[0].tModel, tlp )
		glpopmatrix()

		if swing_stage = 0 then
			iMatrix = plr.object[0].bMatrix.Inverse()
			tLp = light.position*iMatrix
			glpushmatrix()
			glMultMatrixf( plr.object[0].bMatrix )
			do_shadow_pass( tModel(1), tlp )
			glpopmatrix()

		elseif swing_stage = 2 then
			for i as single = 0 to 1 step 1/11
				aMatrix.pointat( spline_point(spline(), i), spline_point(spline(), i+.1) )
				iMatrix = aMatrix.Inverse()
				tLp = light.position*iMatrix
				glpushmatrix()
				glMultMatrixf( aMatrix )
				do_shadow_pass( tModel(3), tlp )
				glpopmatrix()
			next

			iMatrix = plr.object[0].tMatrix.Inverse()
			tLp = light.position*iMatrix
			glpushmatrix()
			glMultMatrixf( plr.object[0].tMatrix )
			do_shadow_pass( tModel(2), tlp )
			glpopmatrix()

		else
			iMatrix = plr.object[0].tMatrix.Inverse()
			tLp = light.position*iMatrix
			glpushmatrix()
			glMultMatrixf( plr.object[0].tMatrix )
			do_shadow_pass( tModel(2), tlp )
			glpopmatrix()

		end if
		glCallList( display.Stencil_Cube )
		'#endif


		'all ortho form here to the end of the loop

		glDisable( GL_LIGHTING )
		glDisable( GL_LIGHT0 )
		glDisable( GL_DEPTH_TEST )

		if swing_stage>0 then

			set_ortho( display.w, display.h )
			glPushMatrix()
			gltranslatef( display.w2, display.h2-ratio*100,0)
			glScaleF(ratio, ratio, 1)

			glEnable( GL_TEXTURE_2D )
			glEnable( GL_BLEND )
			glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA )


			if swing_stage = 1 then
				glColor4f(1,1,1,.75)
				'aim display....
				glBindTexture( GL_TEXTURE_2D, texture(texAim).img )
				glBegin( GL_QUADS )
				glTexCoord2f(0,1)
				glVertex2f(-50,50)
				glTexCoord2f(0,0)
				glVertex2f(-50,-50)
				glTexCoord2f(1,0)
				glVertex2f(50,-50)
				glTexCoord2f(1,1)
				glVertex2f(50,50)
				glEnd()

				'aim cursor....
				glDisable( GL_TEXTURE_2D )
				glLineWidth(5)
				glColor4f(0,0,1,.5)
				glBegin( GL_LINES )
				glVertex2f( -50, vertical*50 )
				glVertex2f( 50, vertical*50 )
				glEnd()

				glColor4f(0,0,1,.5)
				glBegin( GL_LINES )
				glVertex2f( horizontal*50, -50 )
				glVertex2f( horizontal*50, 50 )
				glEnd()
				glEnable( GL_TEXTURE_2D )

			elseif swing_stage = 2 then
				glColor4f(1,1,1,.75)
				'power display.....
				glBindTexture( GL_TEXTURE_2D, texture(texPower).img )
				glBegin( GL_QUADS )
				glTexCoord2f(0,1)
				glVertex2f(-50,50)
				glTexCoord2f(0,0)
				glVertex2f(-50,-50)
				glTexCoord2f(1,0)
				glVertex2f(50,-50)
				glTexCoord2f(1,1)
				glVertex2f(50,50)
				glEnd()


				'power cursor.....
				glBindTexture( GL_TEXTURE_2D, texture(texPowerCursor).img )
				glcolor4f(1,1,1,.5)
				'for i as integer = 0 to impulse_strength
				glBegin( GL_QUADS )
				glTexCoord2f(0,1)
				glVertex2f(-10,(impulse_strength*2)-50)
				glTexCoord2f(0,0)
				glVertex2f(-10,-48)
				glTexCoord2f(1,0)
				glVertex2f(10,-48)
				glTexCoord2f(1,1)
				glVertex2f(10,(impulse_strength*2)-50)
				glEnd()
				'next
			end if


			glPopMatrix()
			drop_ortho()
			glEnable( GL_DEPTH_TEST )

		end if


		myFont.oBegin()
		myFont.oPrint( "contact surface "& plr.object[0].contact_string, 0, display.h-20, 1,1,1, 20 )
		myFont.oPrint( "impulse strength "& impulse_strength, 0, display.h-40, 1,1,1, 20 )
		myFont.oPrint( "swing stage "& swing_stage, 0, display.h-60, 1,1,1, 20 )

		if plr.object[0].foul <>0 then
			'myFont.print display.w2-len(print_foul)*8,display.h2, print_foul
			myFont.oPrint( print_foul, display.w2-len(print_foul)*12, display.h2-20, 1,0,0, 40 )
		end if

		if plr.object[0].hit_goal <>0 then
			myFont.oPrint( "Checking...", display.w2-len("Checking...")*8, display.h2-10, 1,1,0, 20 )
		end if

		if plr.object[0].win <>0 then

			if plr.object[0].shots = 1 then
				myFont.oPrint( "HOLE IN ONE!!!", display.w2-len("HOLE IN ONE!!!")*12, display.h2-20, rnd,rnd,rnd, 40 )
			else
				myFont.oPrint( "GOOD JOB", display.w2-len("GOOD JOB")*12, display.h2-20, rnd,rnd,rnd, 40 )
			end if

			if timer>plr.object[0].win_time then
				plr.object[0].win = 0
				load_level( "course2.lwo", nWorld, world, plr, cam, tModel(0), texture() )
				vcam.p = cam.object[0].tMatrix.Position
				tpos = find_surface_center( tModel(0), "tee box" )
				gpos = find_surface_center( tModel(0), "goal" )
				fly_through = true
				fly_toggle = 0
				fly_count = 0
			end if

		end if

		myFont.oEnd()

		flip

		sleep 3,1

	loop until multikey(FB.SC_ESCAPE)


	if plr.object then
		if plr.joint then
			plr.max_joints = 0
			deallocate(plr.joint)
		end if
		deallocate( plr.object )
	end if

	if cam.object then
		deallocate( cam.object )
	end if

	if world.object then
		deallocate( world.object )
	end if

	h_table->hkill()

	for i as integer = 0 to ubound(tmodel)
		unload_model3d( tmodel(i) )
	next

	for i as integer = 0 to ubound(texture)
		glDeleteTextures( 1, @texture(i).img )
	next

	newtonDestroy( nWorld )

	print #50, "OK!"

	sleep 1000, 1

	close #50

	return 0

end function



sub resort_model_surfaces( model as model_struct ptr )

	for s1 as integer = 0 to Model->max_surfaces - 1
		for s2 as integer = s1 to Model->max_surfaces - 1
			if s1<>s2 then
				if Model->surfaces(s1).RGBAlpha.a <= Model->surfaces(s2).RGBAlpha.a then
					swap Model->surfaces(s1), Model->surfaces(s2)
				end if
			end if
		next
	next
	
	for s1 as integer = 0 to Model->max_surfaces - 1
		print #50, Model->surfaces(s1).surfName & " " & Model->surfaces(s1).RGBAlpha.a
	next

end sub


function find_surface_center( model as model_struct ptr, surfname as string ) as vec3f

	dim as single hits
	dim as vec3f center

	for s as integer = 0 to Model->max_surfaces - 1

		if ltrim(rtrim(model->surfaces(s).surfName)) = surfname then

			for t as integer = Model->surfaces(s).Start_ID to Model->surfaces(s).End_ID

				for p as integer = 1 to 3
					center+=Model->vertices(Model->triangles(t).point_id(p))
					hits+=1
				next

			next

		end if

	next

	center/=hits

	return center

end function



sub prep_1st_pass(byref light as light_struct)

	dim as rgbalpha tamb = type( .4,.4,.4,1)
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @tamb.r)
	glEnable(GL_DEPTH_TEST)
	glDepthFunc(GL_LESS)
	glEnable(GL_CULL_FACE)
	glCullFace(GL_BACK)
	glEnable(GL_LIGHTING)
	glDisable(GL_LIGHT0)

end sub

sub render_1st_pass( byref model as model_struct ptr, byref tMatrix as matrix, texture() as texture_struct )
	
	'glPushAttrib( GL_ENABLE_BIT )
	glPushMatrix()
	glMultMatrixf( tMatrix )
	
	'glColor4f(1,1,1,.1)
	for s as integer = 0 to model->max_surfaces-1
		glcolor4fv( @Model->surfaces(s).RGBAlpha.R )
		glCallList( model->surfaces(s).dList )
		'render_model_surface( model, texture(), s  )
	next
	
	'render_model( model, texture() )
	'glCallList( model->dList )
	glPopMatrix()
	'glPopAttrib()
end sub


sub prep_2nd_pass()

	glDepthMask( 0 )
	glEnable( GL_BLEND )
	glBlendFunc( GL_ONE, GL_ONE )
	dim as rgbalpha zamb = type( 0f,0f,0f,1f )
	glLightModelfv( GL_LIGHT_MODEL_AMBIENT, @zamb.r )

	glClear( GL_STENCIL_BUFFER_BIT )
	glColorMask( 0, 0, 0, 0 )
	glEnable( GL_STENCIL_TEST )
	glStencilFunc( GL_ALWAYS, 0, not 0 )
	glStencilMask( not 0 )

end sub

sub render_2nd_pass( byref model as model_struct ptr,  byref tMatrix as matrix, byref light as light_struct )

	if model->shadow_caster <> 0 then

		dim as matrix imat = tMatrix.Inverse()
		dim as vec4f tLp = light.position * imat

		glPushMatrix()
		glMultMatrixf( tMatrix )
		calc_shadow_volume_wCaps( model, tLp )
		do_shadow_pass_wCaps( model, tLp )
		glPopMatrix()

	end if

end sub


sub prep_3rd_pass( byref light as light_struct )

	glStencilFunc( GL_EQUAL, 0, not 0 )
	glStencilOp( GL_KEEP, GL_KEEP, GL_INCR )
	glDepthFunc( GL_EQUAL )
	glColorMask( 1,1,1,1 )

	glEnable( GL_LIGHT0 )
	glEnable( GL_LIGHTING )

	'dim as rgbalpha tamb = type( .2, .2, .2, 1.0 )
	glLightModelfv( GL_LIGHT_MODEL_AMBIENT, @light.ambient.r )
	gllightfv( GL_LIGHT0, GL_POSITION, @light.position.x )
	glLightfv( GL_LIGHT0, GL_AMBIENT, @light.ambient.r )
	glLightfv( GL_LIGHT0, GL_DIFFUSE, @light.diffuse.r )
	glLightfv( GL_LIGHT0, GL_SPECULAR, @light.specular.r )

end sub

sub render_3rd_pass( byref model as model_struct ptr, byref tMatrix as matrix, texture() as texture_struct  )

	glPushMatrix()
	glMultMatrixf( tMatrix )
	render_model( model, texture() )
	glPopMatrix()

end sub


sub load_shaders()

	'temporary sub to load default shaders

	if use_shaders and _shader100_ then
		dim as gluint Shader_Compile_Success
		dim as GlHandleARB Vertex_Bump, Fragment_Bump

		Vertex_Bump   = init_shader( "res/shaders/bump4.vert", GL_VERTEX_SHADER_ARB )
		Fragment_Bump = init_shader( "res/shaders/bump4.frag", GL_FRAGMENT_SHADER_ARB )

		Shader_Bump = GlCreateProgramObjectARB()
		glAttachObjectARB( Shader_Bump, Vertex_Bump )
		glAttachObjectARB( Shader_Bump, Fragment_Bump )
		glLinkProgramARB( Shader_Bump )

		GlValidateProgramARB( Shader_Bump )
		glGetObjectParameterivARB( Shader_Bump, GL_OBJECT_VALIDATE_STATUS_ARB, @Shader_Compile_Success )
		if Shader_Compile_Success = 0 then
			print #50, "GLSL program(bump4) failed to compile."
		else
			print #50, "GLSL program(bump4) compiled successfully."
		end if

		'This gets the memory location of variable("name"), so that we can send data to the gpu at the correct address.
		DecalLoc        = glGetUniformLocationARB( Shader_Bump, strptr("colorMap") )
		NormalLoc       = glGetUniformLocationARB( Shader_Bump, strptr("normalMap") )

		glUseProgramObjectARB( Shader_Bump )
		glUniform1iARB(DecalLoc, 0 )
		glUniform1iARB(NormalLoc, 1 )
		glUseProgramObjectARB( 0 )

		glDeleteObjectArb( Vertex_Bump )
		glDeleteObjectArb( Fragment_Bump )


		'shader 2_____________________________________________________________________

		Vertex_Bump   = init_shader( "res/shaders/multi_tex_bump.vert", GL_VERTEX_SHADER_ARB )
		Fragment_Bump = init_shader( "res/shaders/multi_tex_bump.frag", GL_FRAGMENT_SHADER_ARB )

		Shader_MultiBump = GlCreateProgramObjectARB()
		glAttachObjectARB( Shader_MultiBump, Vertex_Bump )
		glAttachObjectARB( Shader_MultiBump, Fragment_Bump )
		glLinkProgramARB( Shader_MultiBump )

		GlValidateProgramARB( Shader_MultiBump )
		glGetObjectParameterivARB( Shader_MultiBump, GL_OBJECT_VALIDATE_STATUS_ARB, @Shader_Compile_Success )
		if Shader_Compile_Success = 0 then
			print #50, "GLSL program(multi_tex_bump) failed to compile."
		else
			print #50, "GLSL program(multi_tex_bump) compiled successfully."
		end if

		colormap1        = glGetUniformLocationARB( Shader_MultiBump, strptr("colorMap") )
		colormap2        = glGetUniformLocationARB( Shader_MultiBump, strptr("colorMap2") )
		alphamap 		 = glGetUniformLocationARB( Shader_MultiBump, strptr("alphaMap") )
		normalmap 		 = glGetUniformLocationARB( Shader_MultiBump, strptr("normalMap") )

		glUseProgramObjectARB( Shader_MultiBump )
		glUniform1iARB(alphamap, 0 )
		glUniform1iARB(colormap1, 1 )
		glUniform1iARB(colormap2, 2 )
		glUniform1iARB(normalmap, 3 )
		glUseProgramObjectARB( 0 )

		glDeleteObjectArb( Vertex_Bump )
		glDeleteObjectArb( Fragment_Bump )


		'shader 3 toon

		Vertex_Bump   = init_shader( "res/shaders/toon.vert", GL_VERTEX_SHADER_ARB )
		Fragment_Bump = init_shader( "res/shaders/toon.frag", GL_FRAGMENT_SHADER_ARB )

		Shader_toon = GlCreateProgramObjectARB()
		glAttachObjectARB( Shader_toon, Vertex_Bump )
		glAttachObjectARB( Shader_toon, Fragment_Bump )
		glLinkProgramARB( Shader_toon )

		GlValidateProgramARB( Shader_toon )
		glGetObjectParameterivARB( Shader_toon, GL_OBJECT_VALIDATE_STATUS_ARB, @Shader_Compile_Success )
		if Shader_Compile_Success = 0 then
			print #50, "GLSL program(toon) failed to compile."
		else
			print #50, "GLSL program(toon) compiled successfully."
		end if

		myText        	= glGetUniformLocationARB( Shader_toon, strptr("myText") )
		alphaTex 		= glGetUniformLocationARB( Shader_toon, strptr("alphaText") )
		enable_texture = glGetUniformLocationARB( Shader_toon, strptr("enable_texture") )

		glUseProgramObjectARB( Shader_toon )
		glUniform1iARB(myText, 0 )
		glUniform1iARB(alphaTex, 1 )
		glUniform1iARB(enable_texture, 1 )
		glUseProgramObjectARB( 0 )

		glDeleteObjectArb( Vertex_Bump )
		glDeleteObjectArb( Fragment_Bump )

	end if

end sub


Function Spline_Point( Points() as vec3f, byval p as single ) as vec3f

	dim as vec3f rPoint
	dim as vec3f a,b,c

	c.x = 3 * (Points(1).X - Points(0).X)
	b.x = 3 * (Points(3).X - Points(1).X) - c.x
	a.x = Points(2).X - Points(0).X - c.x - b.x

	c.y = 3 * (Points(1).Y - Points(0).Y)
	b.y = 3 * (Points(3).Y - Points(1).Y) - c.y
	a.y = Points(2).Y - Points(0).Y - c.y - b.y

	c.z = 3 * (Points(1).Z - Points(0).Z)
	b.z = 3 * (Points(3).Z - Points(1).Z) - c.z
	a.z = Points(2).Z - Points(0).Z - c.z - b.z

	rPoint.X = a.x * p ^ 3 + b.x * p ^ 2 + c.x * p + Points(0).X
	rPoint.Y = a.y * p ^ 3 + b.y * p ^ 2 + c.y * p + Points(0).Y
	rPoint.Z = a.z * p ^ 3 + b.z * p ^ 2 + c.z * p + Points(0).Z

	return rPoint

end function


